
// gdi32.lib is required for linking GetTextDimensions.obj.
// The file is available as part of the Masm32 package from
// http://www.masm32.com/masmdl.htm

#pragma library("gdi32.lib")

#INCLUDE "dll.CH"
#INCLUDE "common.CH"
#INCLUDE "gra.CH"
#INCLUDE "xbp.CH"

/*

Purposes:

Introduce standard WinAPI graphics functions for speed tuning and enhanced functionality
in Xbase++ .

Demonstrate the oPS:lockHDC() and oPS:UnlockHDC() methods of the XbpPresSpace object
(new for version 1.9.x ), which give access to an initialized Windows device context,
a prerequisite for using native Windows graphics functions.

Share a work-around for 1.8 (and possibly earlier) Xbase++ versions to emulate
oPS:lockHDC() and oPS:UnlockHDC().

Demonstrate the WinAPI function GetTextExtentPoint32, which may be used as a faster replacement
for GraQueryTextBox in some of its most frequent use cases. (It is no plug-in replacement though,
since parameters and return types are incompatible.)

Demonstrate the WinAPI function GetTextExtentExPoint, which is useful for fast word-wrapping
calculations.

Finally show how using the Xbase++ C API for selected graphics operations
may increase speed even more dramatically.

Special dependencies:  gdi32.lib
                       That file is available as part of the Masm32 package from
                       http://www.masm32.com/masmdl.htm

July 29, 2005

Author: Gnter Beyes, XSYS Datenkommunikation GmbH

*/

PROCEDURE Main()

LOCAL cCaption := "GraQueryTextBox vs. GetTextExtentPoint32, Xpp version " + ltrim(str(XPPVER))
LOCAL oPS
LOCAL oFont
LOCAL hDC, nHandle, aHandles
LOCAL i, j, nLen
LOCAL nWidth := 0, nHeight := 0
LOCAL aWidthXB, aWidthAPI
LOCAL nStart, nDurationXB, nDurationDllCall, nDurationC_API, nDurationFit
LOCAL aPoints
LOCAL cText, cMsg
LOCAL nMaxFit
LOCAL aPartialWidths
LOCAL aLines
LOCAL nMax
LOCAL nLaps := 20

MsgBox( "Click 'OK' to start timing now", cCaption )

// Test string

cText := "There are two general types of key-based algorithms: symmetric and "    + ;
   "public-key. Symmetric algorithms, sometimes called conventional "             + ;
   "algorithms, are algorithms where the encryption key can be calculated "       + ;
   "from the decryption key and vice versa. In most symmetric algorithms, the "  + ;
   "encryption key and the decryption key are the same. These algorithms, also "  + ;
   "called secret-key algorithms, single-key algorithms, or one-key "             + ;
   "algorithms, require that the sender and receiver agree on a key before "      + ;
   "they can communicate securely. The security of a symmetric algorithm rests "  + ;
   "in the key; divulging the key means that anyone could encrypt and decrypt "   + ;
   "messages. As long as the communication needs to remain secret, the key "      + ;
   "must remain secret."

// set up presentation space for the default printer

oPS := PrinterPS()

oFont := XbpFont():New( oPS )
oFont:Create("15.Tahoma")

oPS:setFont( oFont )

// ---------------------------------------------------

nLen := len( cText )

// arrays to hold calculated text widths (for comparison)

aWidthXB  := array( nLen)
aWidthAPI := array( nLen )

// do some laps for easier timing

// GraQueryTextBox() version
nStart := seconds()
FOR j := 1 TO nLaps
   FOR i := 1 TO nLen
      aPoints     :=  GraQueryTextBox( oPS, substr( cText, 1, i ) )
      aWidthXB[i] :=  aPoints[3][1]  - aPoints[2][1]
   NEXT
NEXT

nDurationXB := seconds() - nStart

/*
Create a WinAPI device context with correct font selected.

For Xbase++ 1.9.x, use the new method oPS:lockHDC().

For earlier versions, use oPS:device():getHDC() to obtain
an uninitialized WinAPI printer device context, create a WinAPI
clone of the XbpFont object assigned to the printer's presentation space,
and attach that WinAPI font to the WinAPI device context.

*/

#if XPPVER > 1900000
   nHandle := oPS:lockHDC( 4, @hDC ) // param value 4: select font into device context.
#else
   aHandles := LockHDC( oPS )
   hDC      := aHandles[1]
#endif

// Next timing test:
// WinAPI-based via DllExecuteCall
nStart := seconds()
FOR j := 1 TO nLaps
   FOR i := 1 TO nLen
      GetTextExtents( hDC, substr( cText, 1, i ), @nWidth, @nHeight )
      aWidthAPI[i] := nWidth
   NEXT
NEXT
nDurationDllCall := seconds() - nStart

// Next timing test:
// WinAPI-based via C-API
nStart := seconds()
FOR j := 1 TO nLaps
   FOR i := 1 TO nLen
      _GetTextDimensions( hDC, substr( cText, 1, i ), @nWidth, @nHeight )
      aWidthAPI[i] := nWidth
   NEXT
NEXT
nDurationC_API := seconds() - nStart

// Next timing test:
// Determine lines to fit within 1000 device units on the printer's presentation space.
// This might be sped up significantly by allocating fixed memory
// for cText and use pointers instead of substr().

nStart := seconds()
FOR j := 1 TO nLaps
   aPartialWidths := {}
   aLines         := {}
   cMsg           := cText
   DO WHILE len( cMsg ) > 0
      GetTextExtentsEx( hDC, cMsg, NIL, 1000, @nMaxFit, @nWidth, @nHeight, aPartialWidths )
      aadd( aLines, ltrim( left( cMsg, nMaxFit ) ) )
      cMsg := substr( cMsg, nMaxFit+1 )
   ENDDO
NEXT
nDurationFit := seconds() - nStart

// clean up

#if XPPVER > 1900000
   oPS:UnlockHDC( nHandle, hDC )
#else
   UnlockHDC( aHandles )
#endif

oPS:destroy()

// report results

cMsg := "Performed " + ltrim(str( nLaps * nLen )) + " calls to determine string width." + chr(13) + chr(13)

cMsg += ;
   "1. Timing GraQueryTextBox()"   + chr(9)+ chr(9) + ltrim(str(nDurationXB)) + " s" + chr(13) + ;
   "2. Timing GetTextExtents()"    + chr(9)+ chr(9) + ltrim(str(nDurationDllCall)) + " s" + chr(13) + ;
   "3. Timing GetTextDimensions()" + chr(9) + ltrim(str(nDurationC_API)) + " s" + chr(13) + chr(13) + ;
   "2 versus 1: " + ltrim(str( nDurationXB/nDurationDllCall )) + chr(13) + ;
   "3 versus 1: " + ltrim(str( nDurationXB/nDurationC_API ))

MsgBox( cMsg, cCaption )

// ---------------

cMsg := "And this is our test text, split to fit" + chr(13)+;
        "within an area of 1000 device units width" + chr(13)+;
        "on the systems's default printer." + chr(13)+;
        "(Without attempting word break or even hyphenation.)" + chr(13)+;
        "Timing for " + ltrim(str(nLaps)) + " laps: " + ltrim(str(nDurationFit)) + " s" + ;
        chr(13)+chr(13)

nMax := max( 10, len( aLines ) )

FOR i := 1 TO nMax
   cMsg += aLines[i]
   IF i < nMax
      cMsg += chr(13)
   ENDIF
NEXT

MsgBox( cMsg, "Checking out GetTextExtentsEx" )

RETURN

proc appsys() ; return
proc dbesys() ; return

// --------------------------------------------------------
/*
Create a presentation space for the default printer.
*/

FUNCTION PrinterPS()
   LOCAL oPS, oPrinter := XbpPrinter():New()
   LOCAL aSize

   oPrinter:Create()

   oPS   := XbpPresSpace():New()

   aSize := oPrinter:paperSize()

   // size of printable area
   aSize := { aSize[5] - aSize[3], ;
              aSize[6] - aSize[4]  }

   oPS:Create( oPrinter, aSize, GRA_PU_LOMETRIC )

RETURN oPS

// ---------------------------------------------------------

#DEFINE arraybufferstep 50
/*
May be used to split a given string of words separated by chr(32).
Returns an array of integers identifying word breaks.
*/
function GetWordBreaks( cText )

local aBreaks := array( arraybufferstep )
local i       := 0
local nPos

nPos := at( " ", cText )
do while nPos > 0
   IF ++i > len( aBreaks )
      asize( aBreaks, len( aBreaks ) + arraybufferstep )
   ENDIF
   aBreaks[i] := nPos
   nPos := at( " ", cText, nPos+1 )
enddo
asize( aBreaks, i )

return aBreaks

// ---------------------------------------------------------------
// WinAPI stuff
// ---------------------------------------------------------------

/*
Attach a temporary font to the device context returned by oPrinter:getHDC().
That temporary font is an API clone of the current printer font.
Returns the printer device context handle and the handle of its previously attached font.
*/

FUNCTION LockHDC( oPS )

LOCAL oFont    := oPS:setFont()
LOCAL hDC      := oPS:device():getHDC()
LOCAL hFont    := HandleFromFont( oFont )
LOCAL hSavFont := SelectObject( hDC, hFont )

RETURN { hDC, hSavFont }

// --------------------------------------------------
/*
Restore the device context's previous font
before destroying both the temporary font and the device context.
*/
FUNCTION UnlockHDC( aHandles )

LOCAL lOK
LOCAL hDC      := aHandles[1]
LOCAL hSavFont := aHandles[2]
LOCAL hFont    := SelectObject( hDC, hSavFont )

lOK := DeleteObject( hFont )
lOK := DeleteDC( hDC ) .AND. lOK

RETURN lOK

// -------------------------------------------------
// some #DEFINEs for the CreateFont() API function.

#define LOGPIXELSY          90

#define FW_NORMAL          400
#define FW_BOLD            700

#define OUT_DEFAULT_PRECIS   0
#define OUT_TT_PRECIS        4
#define OUT_DEVICE_PRECIS    5
#define OUT_RASTER_PRECIS    6

#define DEFAULT_PITCH        0
#define FIXED_PITCH          1
#define VARIABLE_PITCH       2

#define CLIP_DEFAULT_PRECIS     0
#define CLIP_CHARACTER_PRECIS   1
#define CLIP_STROKE_PRECIS      2

#define DEFAULT_QUALITY         0
#define DRAFT_QUALITY           1
#define PROOF_QUALITY           2
#define NONANTIALIASED_QUALITY  3
#define ANTIALIASED_QUALITY     4

/*
In the absence of an XbpFont:getHandle() method,
this function attempts to create a clone of an XbpFont object using WinAPI,
returning a font handle.
*/

FUNCTION HandleFromFont( oFont )

LOCAL nHeight
LOCAL nWidth
LOCAL nEscapement
LOCAL nOrientation
LOCAL fnWeight
LOCAL fdwItalic
LOCAL fdwUnderline
LOCAL fdwStrikeOut
LOCAL fdwCharSet
LOCAL fdwOutputPrecision
LOCAL fdwClipPrecision
LOCAL fdwQuality
LOCAL fdwPitchAndFamily
LOCAL cFaceName
LOCAL nLogPixelsY

   nHeight            := oFont:Height
   nWidth             := 0             // let the GDI font mapper figure that one out

   // For the following six font parameters,
   // there is no, or no complete, match in XbpFont IVars, unfortunately;
   // so they should be guessed, if at all possible.

   nEscapement        := 0
   nOrientation       := 0
   fdwOutputPrecision := 0
   fdwClipPrecision   := 0
   fdwQuality         := 0  // partly available via oFont:antiAliased
   fdwPitchAndFamily  := 0  // pitch is available via oFont:Fixed, font family is N/A

   fnWeight           := oFont:WeightClass
   IF fnWeight = 0
      IF oFont:bold
         fnWeight := FW_BOLD
      ELSE
         fnWeight := FW_NORMAL
      ENDIF
   ENDIF
   fdwItalic          := IIF( oFont:italic, 1, 0 )
   fdwUnderline       := IIF( oFont:underscore, 1, 0 )
   fdwStrikeOut       := IIF( oFont:strikeOut, 1, 0 )
   fdwCharset         := oFont:codePage

   IF oFont:status() = XBP_STAT_CREATE
      // oFont:generic is only available after oFont:create().
      IF ! oFont:generic
         fdwOutputPrecision := OUT_DEVICE_PRECIS  // device font
      ELSEIF oFont:vector
         fdwOutputPrecision := OUT_TT_PRECIS      // TrueType font
      ELSE
         fdwOutputPrecision := OUT_RASTER_PRECIS  // Bitmap font
      ENDIF
   ELSE
      IF oFont:vector
         fdwOutputPrecision := OUT_TT_PRECIS      // TrueType font
      ELSE
         fdwOutputPrecision := OUT_DEFAULT_PRECIS // let the GDI font mapper decide
      ENDIF
   ENDIF

   IF oFont:Fixed
      fdwPitchAndFamily  := FIXED_PITCH
   ELSE
      fdwPitchAndFamily  := VARIABLE_PITCH
   ENDIF

   fdwClipPrecision   := CLIP_STROKE_PRECIS
   fdwQuality         := DRAFT_QUALITY
   IF oFont:antiAliased
      fdwQuality      := ANTIALIASED_QUALITY
   ENDIF

   cFaceName          := oFont:familyName

RETURN CreateFont( nHeight, nWidth, nEscapement, nOrientation,;
                   fnWeight, fdwItalic, fdwUnderline, fdwStrikeOut, fdwCharSet, ;
                   fdwOutputPrecision, fdwClipPrecision, fdwQuality, ;
                   fdwPitchAndFamily, cFaceName ;
                 )

// ---------------------------------------------------------------
// WinAPI DllCalls
// ---------------------------------------------------------------

/*
Function GetTextExtentPoint32 from gdi32.dll

C declaration:

BOOL GetTextExtentPoint32(
  HDC hdc,           // handle to DC
  LPCTSTR lpString,  // pointer to text string
  int cbString,      // characters in string
  LPSIZE lpSize      // pointer to string size structure
);

SIZE structure:

typedef struct tagSIZE {
  LONG cx;
  LONG cy;
} SIZE, *PSIZE;

*/

/*
Determine the width and height of string <cText>,
such as it would be rendered in device context <hDC>.
Pass <nWidth> and <nHeight> by reference.

Returns a LOGICAL value indicating success or failure.
*/

FUNCTION GetTextExtents( hDC, cText, nWidth, nHeight )

LOCAL sSize := space(8)
LOCAL lOK   := FALSE

STATIC tpl
IF tpl = NIL
   tpl := DllPrepareCall("gdi32.dll", DLL_STDCALL, "GetTextExtentPoint32A" )
ENDIF

IF DllExecuteCall( tpl, hDC, cText, len( cText ), @sSize ) <> 0
   lOK     := TRUE
   nWidth  := Bin2L( substr( sSize, 1, 4 ) )
   nHeight := Bin2L( substr( sSize, 5, 4 ) )
ENDIF

RETURN lOK

// -------------------------------------------------------------------
/*
Create a font, returning a font handle.
*/

FUNCTION CreateFont( nHeight, nWidth, nEscapement, nOrientation,;
                     fnWeight, fdwItalic, fdwUnderline, fdwStrikeOut, fdwCharSet, ;
                     fdwOutputPrecision, fdwClipPrecision, fdwQuality, ;
                     fdwPitchAndFamily, cFaceName ;
                   )

STATIC tpl
IF tpl == NIL
   tpl := DllPrepareCall( "gdi32.dll", DLL_STDCALL, "CreateFontA" )
ENDIF

RETURN DllExecuteCall( tpl,;
                       nHeight, nWidth, nEscapement, nOrientation,;
                       fnWeight, fdwItalic, fdwUnderline, fdwStrikeOut, fdwCharSet, ;
                       fdwOutputPrecision, fdwClipPrecision, fdwQuality, ;
                       fdwPitchAndFamily, cFaceName ;
                     )
// ----------------------------------------------
/*
Attach a Windows graphics object <hObject>, such as a font, to a device context <hDC>.
Returns the previously attached Windows graphics object of the same type,
e.g. the previously attached font.
*/

FUNCTION SelectObject( hDC, hObject )

STATIC tpl
IF tpl == NIL
   tpl := DllPrepareCall( "gdi32.dll", DLL_STDCALL, "SelectObject" )
ENDIF

RETURN DllExecuteCall( tpl, hDC, hObject )

// ----------------------------------------------
/*
Destroy a Windows graphics object, such as a font.
*/

FUNCTION DeleteObject( hObject )

STATIC tpl
IF tpl == NIL
   tpl := DllPrepareCall( "gdi32.dll", DLL_STDCALL, "DeleteObject" )
ENDIF

RETURN ( DllExecuteCall( tpl, hObject ) <> 0 )

// ---------------------------------------------
/*
Destroy a device context created by CreateDC(), e.g. a printer device context.
*/

FUNCTION DeleteDC( hDC )

STATIC tpl
IF tpl == NIL
   tpl := DllPrepareCall( "gdi32.dll", DLL_STDCALL, "DeleteDC" )
ENDIF

RETURN ( DllExecuteCall( tpl, hDC ) <> 0 )
// -----------------------------------------------------

/*
Retrieves the number of characters in <cString>
that will fit within a space specified in <nMaxExtent>,
as well as the actually required width and height.
Optionally fills an array with the text extent for each of those characters.

<nLength> defaults to len( <cString> ).

Pass <nMaxFit>, <nWidth>, <nHeight> by reference.
If an array of text extents for each of those characters is required,
pass an empty array for <aPartialWidths>.

Returns a LOGICAL value indicating success or failure.
*/

FUNCTION GetTextExtentsEx( hDC, cString, nLength, nMaxExtent, nMaxFit, nWidth, nHeight, aPartialWidths )

LOCAL rc, i
LOCAL sExtents
LOCAL sSize := space( 8 )

STATIC tpl
IF tpl == NIL
   tpl := DllPrepareCall( "gdi32.dll", DLL_STDCALL, "GetTextExtentExPointA" )
ENDIF

IF ISCHARACTER( cString )
   DEFAULT nLength TO len( cString )
ENDIF

nMaxFit := 0
nWidth  := 0
nHeight := 0

IF ISARRAY( aPartialWidths )
   sExtents := space( 4 * nLength )
   rc := DllExecuteCall( tpl,;
                         hDC, cString, nLength, nMaxExtent, @nMaxFit, @sExtents, @sSize )

   IF rc > 0
      asize( aPartialWidths, nMaxFit )
      FOR i := 1 TO nMaxFit
         aPartialWidths[i] := Bin2L( substr( sExtents, ((i-1)*4) + 1, 4 ) )
      NEXT
   ENDIF
ELSE
   rc := DllExecuteCall( tpl,;
                         hDC, cString, nLength, nMaxExtent, @nMaxFit, 0, @sSize )

ENDIF

nWidth  := Bin2L( substr( sSize, 1, 4 ) )
nHeight := Bin2L( substr( sSize, 5, 4 ) )

RETURN rc <> 0

// =============================================================================

/*
The encyclopedic MSDN entry for GetTextExtentExPoint is included for reference.

The GetTextExtentExPoint function retrieves the number of characters in a specified string
that will fit within a specified space and fills an array with the text extent
for each of those characters.
(A text extent is the distance between the beginning of the space
and a character that will fit in the space.)
This information is useful for word-wrapping calculations.

BOOL GetTextExtentExPoint(
  HDC hdc,         // handle to DC
  LPCTSTR lpszStr, // character string
  int cchString,   // number of characters
  int nMaxExtent,  // maximum width of formatted string
  LPINT lpnFit,    // maximum number of characters
  LPINT alpDx,     // array of partial string widths
  LPSIZE lpSize    // string dimensions
);
Parameters

hdc
[in] Handle to the device context.

lpszStr
[in] Pointer to the null-terminated string for which extents are to be retrieved.

cchString
[in] Specifies the number of characters in the string pointed to by the lpszStr parameter.
For an ANSI call it specifies the string length in bytes and for a Unicode it specifies
the string length in WORDs.
Note that for the ANSI function, characters in SBCS code pages take one byte each,
while most characters in DBCS code pages take two bytes; for the Unicode function,
most currently defined Unicode characters (those in the Basic Multilingual Plane (BMP))
are one WORD while Unicode surrogates are two WORDs.

Windows 95/98/Me: This value may not exceed 8192.

nMaxExtent
[in] Specifies the maximum allowable width, in logical units, of the formatted string.

lpnFit
[out] Pointer to an integer that receives a count of the maximum number of characters
that will fit in the space specified by the nMaxExtent parameter.
When the lpnFit parameter is NULL, the nMaxExtent parameter is ignored.

alpDx
[out] Pointer to an array of integers that receives partial string extents.
Each element in the array gives the distance, in logical units, between the beginning of the string
and one of the characters that fits in the space specified by the nMaxExtent parameter.
This array must have at least as many elements as characters specified by the cchString parameter
because the entire array is used internally.
The function fills the array with valid extents for as many characters as are specified
by the lpnFit parameter. Any values in the rest of the array should be ignored.
If alpDx is NULL, the function does not compute partial string widths.
For complex scripts, where a sequence of characters may be represented by any number of glyphs,
the values in the alpDx array up to the number specified by the lpnFit parameter match one-to-one
with code points. Again, you should ignore the rest of the values in the alpDx array.

lpSize
[out] Pointer to a SIZE structure that receives the dimensions of the string, in logical units.
This parameter cannot be NULL.

Return Values
If the function succeeds, the return value is nonzero.

If the function fails, the return value is zero.

Windows NT/2000/XP: To get extended error information, call GetLastError.

Remarks
If both the lpnFit and alpDx parameters are NULL, calling the GetTextExtentExPoint function
is equivalent to calling the GetTextExtentPoint function.

For the ANSI version of GetTextExtentExPoint, the lpDx array has the same number of INT values
as there are bytes in lpString. The INT values that correspond to the two bytes of a DBCS character
are each the extent of the entire composite character.

Note, the alpDx values for GetTextExtentExPoint are not the same as the lpDx values for ExtTextOut.
To use the alpDx values in lpDx, you must first process them.

When returning the text extent, this function assumes that the text is horizontal,
that is, that the escapement is always 0.
This is true for both the horizontal and vertical measurements of the text.
Even if using a font specifying a nonzero escapement, this function will not use the angle
while computing the text extent  the application must convert it explicitly.

*/

